# -*- coding: utf8 -*-
import os

from django.conf import settings

SQLIBRIST_DIRECTORY = getattr(settings,
                              'SQLIBRIST_DIRECTORY',
                              os.path.join(settings.BASE_DIR, 'sql'))



# -*- coding: utf8 -*-
from __future__ import print_function

from sqlibrist.helpers import get_engine


def test_connection(args, config):
    engine = get_engine(config)

    connection = engine.get_connection()

    print('Connection OK')
    connection.close()



# -*- coding: utf8 -*-
from __future__ import print_function

from sqlibrist.helpers import get_last_schema, save_migration, \
    get_current_schema, compare_schemas, mark_affected_items


def makemigration(args, config):
    empty = args.empty
    dry_run = args.dry_run
    migration_name = args.name

    current_schema = get_current_schema()
    execution_plan_up = []
    execution_plan_down = []

    if not empty:
        last_schema = get_last_schema() or {}

        added, removed, changed = compare_schemas(last_schema, current_schema)

        added_items = sorted([current_schema[name] for name in added],
                             key=lambda i: i['degree'])

        if added_items:
            print('Creating:')
            for item in added_items:
                print(' %s' % item['name'])

                execution_plan_up.append(item['up'])
                execution_plan_down.append(item['down'])

        for name in changed:
            current_schema[name]['status'] = 'changed'
        for name in changed:
            mark_affected_items(current_schema, name)

        changed_items = sorted([item
                                for item in current_schema.values()
                                if item.get('status') == 'changed'],
                               key=lambda i: i['degree'])

        if changed_items:
            print('Updating:')
            print(' dropping:')
            for item in reversed(changed_items):
                if item['down']:
                    print('  %s' % item['name'])

                if item['name'] in last_schema \
                        and last_schema[item['name']]['down']:
                    execution_plan_up.append(last_schema[item['name']]['down'])
                    execution_plan_down.append(last_schema[item['name']]['up'])
                elif item['name'] not in last_schema and item['name']['down']:
                    execution_plan_up.append(item['name']['down'])

            execution_plan_up.append(
                ['-- ==== Add your instruction here ===='])

            print(' creating:')
            for item in changed_items:
                if item['down']:
                    print('  %s' % item['name'])
                    execution_plan_up.append(item['up'])
                    execution_plan_down.append(item['down'])

        removed_items = sorted(
            [last_schema[name] for name in removed],
            key=lambda i: i['degree'],
            reverse=True)
        if removed_items:
            print('Deleting:')
            for item in removed_items:
                print(' %s' % item['name'])

                execution_plan_up.append(item['down'])
                execution_plan_down.append(last_schema[item['name']]['up'])

        default_suffix = 'auto'
    else:
        default_suffix = 'manual'

    suffix = ('-%s' % (migration_name or default_suffix))

    if not dry_run:
        save_migration(current_schema,
                       execution_plan_up,
                       reversed(execution_plan_down),
                       suffix)



# -*- coding: utf8 -*-
from __future__ import print_function

import argparse
import glob
import hashlib
import os
import re
from json import loads, dumps

from sqlibrist.engines import Postgresql

ENGINE_POSTGRESQL = 'pg'

ENGINES = {
    ENGINE_POSTGRESQL: Postgresql
}


class SqlibristException(Exception):
    pass


class CircularDependencyException(SqlibristException):
    pass


class UnknownDependencyException(SqlibristException):
    pass


class BadConfig(SqlibristException):
    pass


class ApplyMigrationFailed(SqlibristException):
    pass


class MigrationIrreversible(SqlibristException):
    pass


def get_config(args):
    import yaml
    from yaml.scanner import ScannerError

    try:
        with open(args.config_file) as config_file:
            configs = yaml.load(config_file.read())
    except IOError:
        raise BadConfig('No config file %s found!' % args.config_file)
    except ScannerError:
        raise BadConfig('Bad config file syntax')
    else:
        try:
            return configs[args.config]
        except KeyError:
            raise BadConfig('No config named %s found!' % args.config)


def get_engine(config):
    try:
        return ENGINES[config['engine']](config)
    except KeyError:
        raise BadConfig('DB engine not selected in config or wrong engine '
                        'name (must be one of %s)' % ','.join(ENGINES.keys()))


def get_last_schema():
    schemas = sorted(glob.glob('migrations/*'))
    if schemas:
        with open(os.path.join(schemas[-1], 'schema.json'), 'r') as f:
            schema = loads(f.read())
    else:
        schema = {}
    return schema


def extract_reqs(lines):
    for line in lines:
        if line.strip().startswith('--REQ'):
            _, requirement = line.split()
            yield requirement


def extract_up(lines):
    on = False
    for line in lines:
        if line.strip().startswith('--UP'):
            on = True
        elif line.strip().startswith('--DOWN'):
            raise StopIteration
        elif on:
            yield line.rstrip()


def extract_down(lines):
    on = False
    for line in lines:
        if line.strip().startswith('--DOWN'):
            on = True
        elif on:
            yield line.rstrip()


def init_item(directory, filename):
    with open(os.path.join(directory, filename), 'r') as f:
        lines = f.readlines()

    filename = '/'.join(directory.split('/')[1:] + [filename[:-4]])
    requires = list(extract_reqs(lines))
    up = list(extract_up(lines))
    down = list(extract_down(lines))
    _hash = hashlib.md5(re.sub(r'\s{2,}', '', ''.join(up))).hexdigest()

    return (filename,
            {'hash': _hash,
             'name': filename,
             'requires': requires,
             'required': [],
             'up': up,
             'down': down})


def schema_collector():
    files_generator = os.walk('schema')
    for directory, subdirectories, files in files_generator:
        for filename in files:
            if filename.endswith('.sql'):
                yield init_item(directory, filename)


def check_for_circular_dependencies(schema, name, metadata, stack=()):
    if name in stack:
        raise CircularDependencyException(stack + (name,))
    for requires in metadata['requires']:
        check_for_circular_dependencies(schema,
                                        requires,
                                        schema[requires],
                                        stack + (name,))


def calculate_cumulative_degree(schema, name, metadata, degree=0):
    return len(metadata['requires']) \
           + sum([calculate_cumulative_degree(schema,
                                              requirement,
                                              schema[requirement])
                  for requirement in metadata['requires']])


def get_current_schema():
    schema = dict(schema_collector())

    item_names = schema.keys()

    for name, metadata in schema.items():
        for requirement in metadata['requires']:
            if requirement not in item_names:
                raise UnknownDependencyException((requirement, name))

            schema[requirement]['required'].append(name)
    for name, metadata in schema.items():
        check_for_circular_dependencies(schema, name, metadata)
        metadata['degree'] = calculate_cumulative_degree(schema, name, metadata)
    return schema


def compare_schemas(last_schema, current_schema):
    last_set = set(last_schema.keys())
    current_set = set(current_schema.keys())

    added = current_set - last_set
    removed = last_set - current_set
    changed = [item
               for item in last_set.intersection(current_set)
               if last_schema[item]['hash'] != current_schema[item]['hash']]

    return added, removed, changed


def save_migration(schema, plan_up, plan_down, suffix=''):
    migration_name = '%04.f%s' % (len(glob.glob('migrations/*')) + 1, suffix)
    dirname = os.path.join('migrations', migration_name)
    print('Creating new migration %s' % migration_name)
    os.mkdir(dirname)
    schema_filename = os.path.join(dirname, 'schema.json')
    with open(schema_filename, 'w') as f:
        f.write(dumps(schema, indent=2))

    plans = (
        ('up.sql', plan_up),
        ('down.sql', plan_down),
    )
    for plan_name, instructions in plans:
        with open(os.path.join(dirname, plan_name), 'w') as f:
            for item in instructions:
                f.write('-- begin --\n')
                f.write(('\n'.join(map(lambda s: s.encode('utf8'),
                                       item))).strip())
                f.write('\n')
                f.write('-- end --\n')
                f.write('\n\n')


def mark_affected_items(schema, name):
    schema[name]['status'] = 'changed'
    for required in schema[name]['required']:
        mark_affected_items(schema, required)


def handle_exception(e):
    if isinstance(e, CircularDependencyException):
        print('Circular dependency:')
        print('  %s' % ' >\n  '.join(e.message))
    elif isinstance(e, UnknownDependencyException):
        print('Unknown dependency %s at %s' % e.message)
    elif isinstance(e, (BadConfig, MigrationIrreversible)):
        print(e.message)


def get_command_parser(parser=None):
    from sqlibrist.commands.diff import diff
    from sqlibrist.commands.init import init
    from sqlibrist.commands.initdb import initdb
    from sqlibrist.commands.makemigration import makemigration
    from sqlibrist.commands.status import status
    from sqlibrist.commands.test_connection import test_connection
    from sqlibrist.commands.migrate import migrate
    from sqlibrist.commands.info import info

    _parser = parser or argparse.ArgumentParser()
    _parser.add_argument('--config-file', '-f',
                         help='Config file, default is sqlibrist.yaml',
                         type=str,
                         default=os.environ.get('SQLIBRIST_CONFIG_FILE',
                                                'sqlibrist.yaml'))
    _parser.add_argument('--config', '-c',
                         help='Config name in config file, '
                              'default is "default"',
                         type=str,
                         default=os.environ.get('SQLIBRIST_CONFIG', 'default'))

    subparsers = _parser.add_subparsers(parser_class=argparse.ArgumentParser)

    # print info
    print_info_parser = subparsers.add_parser('info',
                                              help='Print sqlibrist info')
    print_info_parser.add_argument('--verbose', '-v',
                                   action='store_true', default=False)
    print_info_parser.set_defaults(func=info)

    # test_connection
    test_connection_parser = subparsers.add_parser('test_connection',
                                                   help='Test DB connection')
    test_connection_parser.add_argument('--verbose', '-v',
                                        action='store_true', default=False)
    test_connection_parser.set_defaults(func=test_connection)

    # init
    init_parser = subparsers.add_parser('init',
                                        help='Init directory structure')
    init_parser.add_argument('--verbose', '-v',
                             action='store_true', default=False)
    init_parser.set_defaults(func=init)

    # initdb
    initdb_parser = subparsers.add_parser('initdb',
                                          help='Create DB table for '
                                               'migrations tracking')
    initdb_parser.add_argument('--verbose', '-v',
                               action='store_true', default=False)
    initdb_parser.set_defaults(func=initdb)

    # makemigrations
    makemigration_parser = subparsers.add_parser('makemigration',
                                                 help='Create new migration')
    makemigration_parser.set_defaults(func=makemigration)
    makemigration_parser.add_argument('--verbose', '-v',
                                      action='store_true', default=False)

    makemigration_parser.add_argument('--empty',
                                      help='Create migration with empty up.sql '
                                           'for manual instructions',
                                      action='store_true',
                                      default=False)
    makemigration_parser.add_argument('--name', '-n',
                                      help='Optional migration name',
                                      type=str,
                                      default='')
    makemigration_parser.add_argument('--dry-run',
                                      help='Do not save migration',
                                      action='store_true',
                                      default=False)

    # migrate
    migrate_parser = subparsers.add_parser('migrate',
                                           help='Apply pending migrations')
    migrate_parser.set_defaults(func=migrate)
    migrate_parser.add_argument('--verbose', '-v',
                                action='store_true', default=False)
    migrate_parser.add_argument('--fake',
                                help='Mark pending migrations as applied',
                                action='store_true',
                                default=False)
    migrate_parser.add_argument('--dry-run',
                                help='Do not make actual changes to the DB',
                                action='store_true',
                                default=False)
    migrate_parser.add_argument('--migration', '-m',
                                help='Apply up to given migration number',
                                type=str)
    migrate_parser.add_argument('--revert', '-r',
                                help='Unapply last migration',
                                action='store_true')

    # diff
    diff_parser = subparsers.add_parser('diff', help='Show changes to schema')
    diff_parser.set_defaults(func=diff)
    diff_parser.add_argument('--verbose', '-v',
                             action='store_true', default=False)

    # status
    status_parser = subparsers.add_parser('status',
                                          help='Show unapplied migrations')
    status_parser.add_argument('--verbose', '-v',
                               action='store_true', default=False)
    status_parser.set_defaults(func=status)
    return _parser



# -*- coding: utf8 -*-
from __future__ import absolute_import

import os
import sys
from contextlib import contextmanager

from django.core.management import BaseCommand

from django_sqlibrist.helpers import get_config
from django_sqlibrist.settings import SQLIBRIST_DIRECTORY
from sqlibrist.helpers import get_command_parser, SqlibristException, \
    handle_exception


@contextmanager
def chdir(target):
    current_dir = os.curdir
    os.chdir(target)
    yield
    os.chdir(current_dir)


class Command(BaseCommand):
    def add_arguments(self, parser):
        get_command_parser(parser)
        self.parser = parser

    def handle(self, *args, **options):
        _args = self.parser.parse_args(sys.argv[2:])

        config = get_config(_args)

        with chdir(SQLIBRIST_DIRECTORY):
            try:
                _args.func(_args, config)
            except SqlibristException as e:
                handle_exception(e)



# -*- coding: utf8 -*-
from django.conf import settings

from sqlibrist.helpers import ENGINE_POSTGRESQL, BadConfig

DEFAULT_PORTS = {
    ENGINE_POSTGRESQL: '5432'
}


def get_config(args):
    DB = settings.DATABASES['default']
    config = {}
    if 'postgresql' in DB.get('ENGINE', '') \
            or 'psycopg' in DB.get('ENGINE', ''):
        config['engine'] = ENGINE_POSTGRESQL
    else:
        raise BadConfig('Django configured with unsupported database engine: '
                        '%s' % DB.get('ENGINE', ''))

    config['name'] = DB.get('NAME', '')
    config['user'] = DB.get('USER', '')
    config['password'] = DB.get('PASSWORD', '')
    config['host'] = DB.get('HOST') or '127.0.0.1'
    config['port'] = DB.get('PORT') or DEFAULT_PORTS[config['engine']]
    return config



# -*- coding: utf8 -*-
VERSION = '0.1.0'

from sqlibrist.helpers import SqlibristException, handle_exception, \
    get_command_parser, get_config


def main():
    parser = get_command_parser()
    args = parser.parse_args()
    config = get_config(args)
    try:
        args.func(args, config)
    except SqlibristException as e:
        handle_exception(e)


if __name__ == '__main__':
    main()



# -*- coding: utf8 -*-
from __future__ import print_function

from sqlibrist.helpers import get_engine


def initdb(args, config):
    engine = get_engine(config)

    print('Creating db...')
    engine.create_migrations_table()
    print('Done.')





# -*- coding: utf8 -*-
from __future__ import print_function

import os

DEFAULT_CONFIG_FILE = 'sqlibrist.yaml'
DEFAULT_CONFIG = """---
default:
  engine: pg
  user: <username>
  name: <database_name>
  password: <password>
# host: 127.0.0.1
# port: 5432
"""


def init(args, config):
    print('Creating directories...')
    dirlist = (
        'schema',
        'schema/tables',
        'schema/functions',
        'schema/views',
        'schema/triggers',
        'schema/indexes',
        'schema/types',
        'schema/constraints',
        'migrations'
    )

    for dirname in dirlist:
        if not os.path.isdir(dirname):
            os.mkdir(dirname)

    if not os.path.isfile(DEFAULT_CONFIG_FILE):
        with open(DEFAULT_CONFIG_FILE, 'w') as f:
            f.write(DEFAULT_CONFIG)
    print('Done.')



# -*- coding: utf8 -*-
from __future__ import absolute_import, print_function


class BaseEngine(object):
    def __init__(self, config):
        self.config = config

    def get_connection(self):
        raise NotImplementedError

    def create_migrations_table(self):
        raise NotImplementedError

    def get_applied_migrations(self):
        raise NotImplementedError

    def apply_migration(self, name, statements, fake=False):
        raise NotImplementedError

    def unapply_migration(self, name, statements, fake=False):
        raise NotImplementedError


class Postgresql(BaseEngine):
    def get_connection(self):
        import psycopg2
        return psycopg2.connect(
            database=self.config.get('name'),
            user=self.config.get('user'),
            host=self.config.get('host'),
            password=self.config.get('password'),
            port=self.config.get('port'),
        )

    def create_migrations_table(self):
        connection = self.get_connection()
        print('Creating schema and migrations log table...\n')
        with connection:
            with connection.cursor() as cursor:
                cursor.execute('CREATE SCHEMA IF NOT EXISTS sqlibrist;')

                cursor.execute('''
                CREATE TABLE IF NOT EXISTS sqlibrist.migrations (
                id SERIAL PRIMARY KEY,
                migration TEXT,
                datetime TIMESTAMPTZ DEFAULT CURRENT_TIMESTAMP
                );
                ''')
        connection.close()

    def get_applied_migrations(self):
        connection = self.get_connection()
        with connection:
            with connection.cursor() as cursor:
                cursor.execute('''
                SELECT migration FROM sqlibrist.migrations
                ORDER BY datetime; ''')
                return cursor.fetchall()

    def get_last_applied_migration(self):
        connection = self.get_connection()
        with connection:
            with connection.cursor() as cursor:
                cursor.execute('''
                SELECT migration FROM sqlibrist.migrations
                ORDER BY datetime DESC
                LIMIT 1; ''')
                result = cursor.fetchone()
                return result and result[0] or None

    def apply_migration(self, name, statements, fake=False):
        import psycopg2
        connection = self.get_connection()
        with connection:
            with connection.cursor() as cursor:
                try:
                    if not fake and statements.strip():
                        cursor.execute(statements)
                except (
                psycopg2.OperationalError, psycopg2.ProgrammingError) as e:
                    connection.rollback()
                    print(e.message)
                    from sqlibrist.helpers import ApplyMigrationFailed

                    raise ApplyMigrationFailed
                else:
                    cursor.execute('INSERT INTO sqlibrist.migrations '
                                   '(migration) VALUES (%s);',
                                   [name.split('/')[-1]])
                    connection.commit()

    def unapply_migration(self, name, statements, fake=False):
        import psycopg2
        connection = self.get_connection()
        with connection:
            with connection.cursor() as cursor:
                try:
                    if not fake:
                        cursor.execute(statements)
                except (
                psycopg2.OperationalError, psycopg2.ProgrammingError) as e:
                    connection.rollback()
                    print(e.message)
                    from sqlibrist.helpers import ApplyMigrationFailed

                    raise ApplyMigrationFailed
                else:
                    cursor.execute('DELETE FROM sqlibrist.migrations '
                                   'WHERE migration = (%s); ', [name])
                    connection.commit()



# -*- coding: utf8 -*-
from __future__ import print_function

import glob
import os

from sqlibrist.helpers import get_engine, ApplyMigrationFailed, \
    MigrationIrreversible


def unapplied_migrations(migration_list, applied_migrations):
    ml = [m.split('/')[-1] for m in migration_list]
    for migration in applied_migrations:
        try:
            ml.remove(migration[0])
        except ValueError:
            print('Miration "%s" is not in created '
                  'migration list, probably, this DB '
                  'is from another branch' % migration[0])
    return ml


def migrate(args, config):
    fake = args.fake
    revert = args.revert
    till_migration_name = args.migration
    engine = get_engine(config)

    applied_migrations = engine.get_applied_migrations()

    if applied_migrations and revert:
        last_applied_migration = applied_migrations[-1]
        try:
            with open(os.path.join('migrations',
                                   last_applied_migration,
                                   'down.sql')) as f:
                down = f.read()
        except IOError:
            raise MigrationIrreversible('Migration %s does not '
                                        'have down.sql - reverting '
                                        'impossible' % last_applied_migration)

        print('Un-Applying migration %s... ' % last_applied_migration, end='')
        if fake:
            print('(fake run) ', end='')
        try:
            engine.unapply_migration(last_applied_migration, down, fake)
        except ApplyMigrationFailed:
            print('Error, rolled back')
        else:
            print('done')
        return

    elif not revert:
        migration_list = unapplied_migrations(
            sorted(glob.glob('migrations/*')),
            applied_migrations)
    else:
        # no migrations at all
        migration_list = sorted(glob.glob('migrations/*'))

    for migration in migration_list:
        with open(os.path.join('migrations', migration, 'up.sql')) as f:
            up = f.read()

        migration_name = migration.split('/')[-1]
        print('Applying migration %s... ' % migration_name, end='')
        if fake:
            print('(fake run) ', end='')
        try:
            engine.apply_migration(migration_name, up, fake)
        except ApplyMigrationFailed:
            print('Error, rolled back')
            break
        else:
            print('done')
        if till_migration_name \
                and migration_name == till_migration_name:
            break
